#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Types used within GPflux (for static type-checking).
"""
from typing import List, Tuple, Union

import tensorflow as tf
import tensorflow_probability as tfp

from gpflow.base import TensorType


def unwrap_dist(dist: tfp.distributions.Distribution) -> tfp.distributions.Distribution:
    """
    Unwrap the given distribution, if it is wrapped in a ``_TensorCoercible``.
    """
    while True:
        inner = getattr(dist, "tensor_distribution", None)
        if inner is None:
            return dist
        dist = inner


ShapeType = Union[tf.TensorShape, List[int], Tuple[int, ...]]
r""" Union of valid types for describing the shape of a `tf.Tensor`\ (-like) object """

ObservationType = List[TensorType]
"""
Type for the ``[inputs, targets]`` list used by :class:`~gpflux.layers.LayerWithObservations`
"""
